%% ordinary least squares, constrained least squares:

m=16; n=8;
A = randn(m,n);
b = randn(m,1);

x_ls = inv(A'*A)*A'*b;  % = A\b

cvx_begin
variable x(n);
minimize( norm(A*x - b, 2) );
cvx_end


l = zeros(n,1);
u = ones(n,1);

cvx_begin
variable x(n);
minimize( norm(A*x - b, 2) );
subject to
x >= l;
x <= u;
cvx_end




%% a generic linear control problem:
nX = 5;
nU = 1;

A = randn(nX, nX);
B = randn(nX, nU);

Q = eye(nX);
R = eye(nU);

x1 = rand(nX,1);

T = 20;

cvx_begin
variables x(5,T)  u(1,T) s(1,T);
minimize( sum(s) )
subject to  
  for t=1:T-1
	  x(:,t+1) == A*x(:,t) + B*u(:,t);
  end
  for t=1:T
	  s(1,t) >= x(:,t)'*Q*x(:,t) + u(:,t)'*R*u(:,t);
  end
  x(:,1) == x1;
cvx_end


figure; plot(x');


% let's execute control policy:

x_run(:,1) = x1;

for t=1:T-1
	x_run(:,t+1) = A*x_run(:,t) + B*u(:,t);
end

figure; plot(x_run');

% let's execute with noise

x_run2(:,1) = x1;

for t=1:T-1
	x_run2(:,t+1) = A*x_run2(:,t) + B*u(:,t) + randn(nX,1);
end

figure; plot(x_run2');


% let's execute a closed loop control policy by using MPC:

x_run3(:,1) = x1;


for k=1:T-1
	cvx_begin
	variables x(nX,T)  u(nU,T) s(1,T);
	minimize( sum(s(k:T)) )
	subject to
	for t=k:T-1
		x(:,t+1) == A*x(:,t) + B*u(:,t);
	end
	for t=k:T
		s(1,t) >= x(:,t)'*Q*x(:,t) + u(:,t)'*R*u(:,t);
	end
	
	x(:,k) == x_run3(:,k);
	cvx_end
	
	x_run3(:,k+1) = A*x_run3(:,k) + B*u(:,k) + randn(nX,1);
	
	

end

figure; plot(x_run3');

%% a nonlinear control problem: cartpole

%% shooting:
clear; clc; close all;

T = 100;
u = randn(1,T)*0.1;
max_iters = 10;
x_init = [-10; 0; 0; 0];
nX = 4; nU = 1;
x_eps = 0.5;
u_eps = 0.5;
dt = 0.1;
Q = eye(nX); R = eye(nU); Q_final = 100*eye(nX);
clear A B c
for iter = 1:max_iters
	% simulate and linearize
	x(:,1) = x_init;
	for t=1:T-1
		x(:,t+1) = sim_cartpole(x(:,t), u(:,t), dt);
		[A{t} B{t} c{t}] = compute_jacobian(@sim_cartpole, x(:,t), u(:,t), dt);
		%cartpole_draw(t*dt, x(:,t));
	end
	figure(1); subplot(3,1,1); hold on; plot(u); ylabel('u'); 
	subplot(3,1,2); hold on; plot(x(1,:)); ylabel('x'); subplot(3,1,3); hold on; plot(x(2,:));ylabel('\theta');
	cost(iter) = 0;
	for t=1:T-1
		cost(iter) = cost(iter) + x(:,t)'*Q*x(:,t) + u(:,t)'*R*u(:,t) + norm(u(:,t+1)-u(:,t),2);
	end
	cost(iter) = cost(iter) + x(:,T)'*Q_final*x(:,T);
	cost
	% solve convex problem
	cvx_begin
	variables x_cvx(nX,T)  u_cvx(nU,T) s_cvx(1,T);
	minimize( sum(s_cvx(1:T)) )
	subject to
	for t=1:T-1
		x_cvx(:,t+1) == A{t}*(x_cvx(:,t)-x(:,t)) + B{t}*(u_cvx(:,t)-u(:,t)) + c{t};
	end
	for t=1:T-1
		s_cvx(1,t) >= x_cvx(:,t)'*Q*x_cvx(:,t) + u_cvx(:,t)'*R*u_cvx(:,t) + norm(u_cvx(:,t+1)-u_cvx(:,t),2);
	end
	s_cvx(1,T) >= x_cvx(:,T)'*Q_final*x_cvx(:,T);
	for t=1:T
		norm(x_cvx(:,t) - x(:,t),2) <= x_eps;
		norm(u_cvx(:,t) - u(:,t),2) <= u_eps;
	end
	x_cvx(:,1) == x_init;
	cvx_end
	u = u_cvx;
end


%% collocation:
clear; clc; close all;

T = 100;
u = randn(1,T)*0.1;
max_iters = 10;
x_init = [-10; pi/10; -0.1; 0.1];
nX = 4; nU = 1;
x_eps = 100;
u_eps = 1;
dt = 0.1;
Q = eye(nX); R = eye(nU); Q_final = 100*eye(nX);
clear A B c
x_target = [0;0;0;0];
for i=1:nX
	x_iter1(i,:) = x_init(i):(x_target(i)-x_init(i))/(T-1):x_target(i);
end
u_iter1=zeros(nU,T);

for iter = 1:max_iters
	% linearize (but no simulation)
	if(iter==1)
		x = x_iter1; u = u_iter1;
	else
		x = x_cvx; u = u_cvx;
	end
	for t=1:T-1
		%	x(:,t+1) = sim_cartpole(x(:,t), u(:,t), dt);
		[A{t} B{t} c{t}] = compute_jacobian(@sim_cartpole, x(:,t), u(:,t), dt);
		%cartpole_draw(t*dt, x(:,t));
	end
	figure(2); subplot(3,1,1); hold on; plot(u); ylabel('u'); 
	subplot(3,1,2); hold on; plot(x(1,:)); ylabel('x'); subplot(3,1,3); hold on; plot(x(2,:));ylabel('\theta');
	cost(iter) = 0;
	for t=1:T-1
		cost(iter) = cost(iter) + x(:,t)'*Q*x(:,t) + u(:,t)'*R*u(:,t) + norm(u(:,t+1)-u(:,t),2);
	end
	cost(iter) = cost(iter) + x(:,T)'*Q_final*x(:,T);
	cost
	% solve convex problem
	cvx_begin
	variables x_cvx(nX,T)  u_cvx(nU,T) s_cvx(1,T);
	minimize( sum(s_cvx(1:T)) )
	subject to
	for t=1:T-1
		x_cvx(:,t+1) == A{t}*(x_cvx(:,t)-x(:,t)) + B{t}*(u_cvx(:,t)-u(:,t)) + c{t};
	end
	for t=1:T-1
		s_cvx(1,t) >= x_cvx(:,t)'*Q*x_cvx(:,t) + u_cvx(:,t)'*R*u_cvx(:,t) + norm(u_cvx(:,t+1)-u_cvx(:,t),2);
	end
	s_cvx(1,T) >= x_cvx(:,T)'*Q_final*x_cvx(:,T);
	for t=1:T
		norm(x_cvx(:,t) - x(:,t),2) <= x_eps;
		norm(u_cvx(:,t) - u(:,t),2) <= u_eps;
	end
	x_cvx(:,1) == x_init;
	cvx_end
	u = u_cvx;
end

% let's evaluate the resulting open-loop sequence:
x(:,1) = x_init;
for t=1:T-1
	x(:,t+1) = sim_cartpole(x(:,t), u(:,t), dt);
end
figure(3); subplot(3,1,1); hold on; plot(u); ylabel('u');
subplot(3,1,2); hold on; plot(x(1,:)); ylabel('x'); subplot(3,1,3); hold on; plot(x(2,:));ylabel('\theta');
final_cost = 0;
for t=1:T-1
  final_cost = final_cost + x(:,t)'*Q*x(:,t) + u(:,t)'*R*u(:,t) + norm(u(:,t+1)-u(:,t),2);
end
final_cost = final_cost + x(:,T)'*Q_final*x(:,T);
final_cost


